/**
 * 
  ASkeyboard.cpp	ASkeyboard-core part of "Hobo-nicola keyboard adapter".
  Copyright (c) 2018 Takeshi Higasa
  
  This file is part of "Hobo-nicola keyboard adapter".

  "Hobo-nicola keyboard adapter" is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  "Hobo-nicola keyboard adapter" is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with "Hobo-nicola keyboard adapter".  If not, see <http://www.gnu.org/licenses/>.
  
  
  ASKeyboard.cpp included in hoboNicola121.zip).  Dec. 1. 2018.
  
*/

#include "ASkeyboard.h"


#define SET_KANA_LED(a)  PORTB = a ? PORTB & ~0x40 : PORTB | 0x40
#define SET_CAPS_LED(a)  PORTE = a ? PORTE & ~0x40 : PORTE | 0x40
#define SET_NUML_LED(a)  PORTC = a ? PORTC & ~0x40 : PORTC | 0x40
#define SET_CXFR_LED(a)  PORTD = a ? PORTD & ~0x80 : PORTD | 0x80
#define SET_KBDL_LED(a)  PORTD = a ? PORTD & ~0x10 : PORTD | 0x10

#define TOGGLE_KANA_LED() (PINB |= 0x40)
#define TOGGLE_CAPS_LED() (PINE |= 0x40)
#define TOGGLE_NUML_LED() (PINC |= 0x40)
#define TOGGLE_CXFR_LED() (PIND |= 0x80)
#define TOGGLE_KBDL_LED() (PIND |= 0x10)


static const uint8_t SCAN_LINES = 15;
static const uint8_t DIP_SW_JIS = 1;
static const uint8_t DIP_SW_OYA1 = 0;
static const uint8_t DIP_SW_OYA2 = 2;


static bool kana_locked;	// ۂScroll lock LED.

static uint8_t scan[SCAN_LINES];
static uint8_t last_scan[SCAN_LINES];
volatile static uint8_t dip_sw;


static const uint8_t KBD_BUFFER_SIZE = 16;
static uint8_t kbd_buffer[KBD_BUFFER_SIZE];  //  fifo key buffer 
static volatile uint8_t kbd_in, kbd_out;

static bool put_buffer(uint8_t s) {
	uint8_t rx = kbd_in + 1;
	if (rx >= KBD_BUFFER_SIZE) 
		rx = 0;
	if (rx == kbd_out) 
		return false; // buffer is full.  
	kbd_buffer[kbd_in] = s;
	kbd_in = rx;    
	return true;  
}

static uint8_t get_buffer() {
	if (kbd_in == kbd_out)
		return 0;  // empty.
	uint8_t s = kbd_buffer[kbd_out];
	cli();
	if (++kbd_out >= KBD_BUFFER_SIZE)
		kbd_out = 0;
	sei();    
	return s;    
}

ASkeyboard::ASkeyboard() {
	memset(scan, 0, sizeof(scan));  
	memset(last_scan, 0, sizeof(last_scan));  
	kbd_led_state = 0;
	modifier_state = last_modifier_state = 0;
	notify = 0;
	fn_key_pressed = false;
	keyboard_locked = false;
	setup_callback = 0;
	kana_locked = false;
}

// VXeNbN16MHzȂ̂ŁA64250KHzB

// 250̃JEgAbv1msecB5msecƂB
static const uint16_t TIMER1_INTERVAL = -250 * 5;	// 0xfb1d

bool ASkeyboard::begin(keyboard_notify* p) {
	notify = p;
	uint8_t oldSREG = SREG;
	cli();
	
	ADCSRA = ADCSRA & 0x7f;
	ACSR = ACSR | 0x80;
	PRR0 = B10000101;
	PRR1 = B00011001;

	// DATA port (input)
	DDRD = DDRD & 0xf0; // PD0-3 as input. (D1-D4)
	DDRF = DDRF & 0x0f; // PF4-7 as input (D5 - D8)
	PIND = PIND | 0x0f;	// INPUT_PULLUP
	PINF = PINF | 0xf0;	// INPUT_PULLUP

	// SCAN port (output)
	DDRB = DDRB |= B00011110; // PB1-4 as OUTPUT
	// LED
	DDRD = DDRD | B10010000; // PD7, PD4 as OUTPUT. C/XFER and KBDLOCK
	DDRC = DDRC | B01000000; // PC6 as OUTPUT. NumLock
	DDRE = DDRE | B01000000; // PE6 as OUTPUT. CapsLock
	DDRB = DDRB | B01000000; // PB6 as OUTPUT. KanaLock

// setup timer1 interrupt (period = 6msec)
	TCCR1A = 0;
	TCCR1B = 3;	// CLKio / 64@ = 250KHz.
	TCCR1C = 0;
	TCNT1 = TIMER1_INTERVAL;
	TIMSK1 = 1;	// enable Overflow Interrupt

	SREG = oldSREG;
	sei();
	led_demo();
	return true;
}

void ASkeyboard::led_demo() {
	SET_KANA_LED(1);
	delay(60);
	SET_KANA_LED(0);
	SET_CAPS_LED(1);
	delay(60);
	SET_CAPS_LED(0);
	SET_CXFR_LED(1);
	delay(60);
	SET_CXFR_LED(0);
	SET_NUML_LED(1);
	delay(60);
	SET_NUML_LED(0);
	SET_KBDL_LED(1);
	delay(60);
	SET_KBDL_LED(0);
}

// timer1 overflow interrupt service routine.
// 5msecƁB
ISR(TIMER1_OVF_vect, ISR_NOBLOCK)
{
	uint8_t tmp_scan[SCAN_LINES];
	for(uint8_t i = 0; i < SCAN_LINES; i++) { // scan all keys. < 200usec.
		PORTB = ((PINB & B11100001) | (i << 1));
		delayMicroseconds(10);    // wait a little.
		tmp_scan[i] = ~((PINF & 0xf0) | (PIND & 0x0f));
	}
	int n = memcmp(tmp_scan, scan, SCAN_LINES);
	memcpy(scan, tmp_scan, SCAN_LINES);
	if (n == 0) {	// OƈvB
// read dip-sw bits and mask them.
		dip_sw = ((scan[10] & 0x20) ? 1 : 0) | ((scan[11] & 0x20) ? 2 : 0);
		scan[10] &= 0xdf;
		scan[11] &= 0xdf;

		uint8_t ok = scan[9] & 3;
		if (!kana_locked) { // OYA1,2ŉpȂΐewL[͋󔒂ɂ
			if (dip_sw == DIP_SW_OYA1 || dip_sw == DIP_SW_OYA2) {
				scan[9] |= ok ? 4 : 0;
				scan[9] &= 0xfc;
			}
		} else if (dip_sw == DIP_SW_OYA2) {	// OYA2łȂȂ΁AewL[͖ϊ^ϊɁB
			scan[9] |= ok << 5;
			scan[9] &= 0xfc;
		}
// generate scan-code.
		for(uint8_t i = 0; i < SCAN_LINES; i++) {
			uint8_t cur = scan[i];
			uint8_t old = last_scan[i];
			uint8_t changed = cur ^ old;
			if (!changed)
			  continue;
			uint8_t mask = 1;
			for(int8_t r = 0; r < 8; r++, mask <<= 1) {
				if (changed & mask) {
					int8_t key = (i << 3) + r + 1;
					key |= (cur & mask) ? 0 : 0x80;	// release Ȃ 0x80OR.
					put_buffer(key);
				}
			}
		}
		memcpy(last_scan, scan, SCAN_LINES);
	}
	TCNT1 = TIMER1_INTERVAL;
}

#define L_OYA	ASKB_LEFT_OYAYUBI
#define R_OYA	ASKB_RIGHT_OYAYUBI

#define L_CTL	0xe0
#define L_SFT	0xe1
#define R_SFT	0xe2
#define L_GUI	0xe3
#define L_ALT	0xe4
#define R_CTL	0xe5

#define FUNC	0

// hid usageid = hid_table[scancode - 1];
static const uint8_t hid_table[] = {
	0x35,  0x65,  0x3a,  0x3b,  0x3c,  0x3d,  0x3e, 0x3f,	// 0x01`0x08
	0x40,  0x41,  0x42,  0x43,  0x29,  0x1e,  0x1f, 0x20,	// 0x09`0x10
	0x21,  0x22,  0x23,  0x24,  0x25,  0x26,  0x27, 0x2d,	// 0x11`0x18
	0x2e,  0x89,  0x14,  0x1a,  0x08,  0x15,  0x17, 0x1c,	// 0x18`0x20
	0x18,  0x0c,  0x12,  0x13,  0x2f,  0x30,  0,    0x04,	// 0x21`0x28
	0x16,  0x07,  0x09,  0x0a,  0x0b,  0x0d,  0x0e, 0x0f,	// 0x29`0x30
	0x33,  0x1d,  0x1b,  0x06,  0x19,  0x05,  0x11, 0x10,	// 0x31`0x38
	0x36,  0x37,  0x38,  0x87,  0x2a,  0x2b,  0x28, L_CTL,	// 0x39`0x40
	0x34,  0x32,  L_SFT, R_SFT, L_GUI, L_ALT, 0x88, 0x39,	// 0x41`0x48
	L_OYA, R_OYA, 0x2c,  R_CTL, FUNC,  0x8b,  0x8a, 0x4b,	// 0x49`0x50
	0x4e,  0x4a,  0x4d,  0x49,  0x4c,  0,     0,    0,		// 0x51`0x58
	0x52,  0x50,  0x4f,  0x51,  0,     0,     0,    0,		// 0x59`0x60
	0x55,  0x57,  0x61,  0x5d,  0x59,  0x85,  0x58, 0,		// 0x61`0x68
	0x54,  0x5f,  0x67,  0x5e,  0x5a,  0x62,  0,    0,		// 0x69`0x70
	0x56,  0x60,  0x5c,  0x63,  0x5b,  0xb1,  0,    0		// 0x71`0x78
};

// HIDƃXLR[hƂ̕ϊe[u
void ASkeyboard::hid_convert(uint8_t k, bool pressed) {
	if (k < 1 || k > 120)
		return;
	uint8_t hid = hid_table[k - 1];
	if ((hid & 0xf0) == 0xe0)	// none-assigned or modifiers.
		return;

	if (fn_key_pressed) {	// key pressed with Fn.
		switch(k) {
		case 0x03:			
			hid = 0x44;		// F11
			break;
		case 0x04:
			hid = 0x45;		//F12
			break;
		case 0x05:
			hid = 0x46;		// F3 --> Print Screen.
			break;
		case 0x06:
			hid = 0x47;		// F4 --> Scrl Lock
			break;
		case 0x07:			
			hid = 0x48;		// F5 --> Pause.
			break;
		case 0x08:			// Alt + F6Őݒ胂[hɓB
			hid = 0;
			if (!keyboard_locked && pressed && setup_callback)
				setup_callback(0);
			break;							
		case 0x0b:
			hid = 0x53;		// F9 -> NumLock.
			break;
		case 0x0c:			// F10 --> Keybaord Lock.
			if (pressed) {
				keyboard_locked ^= 1;
				SET_KBDL_LED(keyboard_locked);
			}
			hid = 0;
			break;
		default:
			hid = 0;
			break;
		}
	}
	if (hid && !keyboard_locked) {
		if (pressed)
			notify->key_pressed(hid, modifier_state);
		else
			notify->key_released(hid, modifier_state);
	}
}

void ASkeyboard::task(uint8_t hid_led) {
	uint8_t key;
	while( (key = get_buffer()) != 0) {
		if (key != 0) {
			bool pressed = true;
			if (key & 0x80) {
				pressed = false;
				key &= 0x7f;
			}
			if (notify) {
				notify->raw_key_event(key, !pressed);
	// modifiers.
				last_modifier_state = modifier_state;
				uint8_t mask = 0;
				switch(key) {
				case ASKB_SHIFT_L:
					mask = MODIFIRE_HID_SHIFT_L;
					break;
				case ASKB_SHIFT_R:
					mask = MODIFIRE_HID_SHIFT_R;
					break;
				case ASKB_CTRL_L:
					mask = MODIFIRE_HID_CTRL_L;
					break;
				case ASKB_CTRL_R:
					mask = MODIFIRE_HID_CTRL_R;
					break;
				case ASKB_ALT_L:
					mask = MODIFIRE_HID_ALT_L;
					break;
				//case ASKB_ALT_R:
				//	mask = MODIFIRE_HID_ALT_R;
				//	break;
				case ASKB_GUI_L:
					mask = MODIFIRE_HID_GUI_L;
					break;
				//case ASKB_GUI_R:
				//	mask = MODIFIRE_HID_GUI_R;
				//	break;
				case ASKB_FN:
					fn_key_pressed = pressed;
					notify->key_released_all();
					break;
				default:
					hid_convert(key, pressed);
					continue;
				}
				if (pressed)
					modifier_state |= mask;
				else
					modifier_state &= ~mask;
				if (!keyboard_locked && modifier_state != last_modifier_state)
					notify->modifier_changed(last_modifier_state, modifier_state);
			}
		}
	}
// hid L[{[hpLED
	SET_NUML_LED((hid_led & KEY_LED_NUMLOCK) ? 1 : 0);
	SET_CAPS_LED((hid_led & KEY_LED_CAPSLOCK) ? 1 : 0);
	kana_locked = (bool)(hid_led & KEY_LED_SCRLOCK);
	SET_KANA_LED(kana_locked);
}

// dip-sw pos == JIS.
bool ASkeyboard::is_nicola_disabled() const {
	return dip_sw == DIP_SW_JIS;
}

void ASkeyboard::setup_led(bool on) {
	SET_CXFR_LED(on);
}

// ewĥƂ́AϊEϊewL[ɂȂB
bool ASkeyboard::dedicated_oyayubi() const {
	return (dip_sw == DIP_SW_OYA1);
}

