/**
 * 
  ASkeyboard.h	Definitions of ASkeyboard class for "Hobo-nicola keyboard adapter".
  Copyright (c) 2018 Takeshi Higasa
  
  This file is part of "Hobo-nicola keyboard adapter".

  "Hobo-nicola keyboard adapter" is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  "Hobo-nicola keyboard adapter" is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with "Hobo-nicola keyboard adapter".  If not, see <http://www.gnu.org/licenses/>.
  
*/
#include <Arduino.h>
#include "hbn_defines.h"

// scan-code
#define ASKB_CTRL_L		0x40
#define ASKB_CTRL_R		0x4c
#define ASKB_SHIFT_L	0x43
#define ASKB_SHIFT_R	0x44
#define ASKB_ALT_L		0x46	// GPRH
#define ASKB_ALT_R
#define ASKB_GUI_L		0x45	// CAPS
#define ASKB_GUI_R

#define ASKB_EISU		0x48	// TAB(NFER)
#define ASKB_HIRAGANA	0x47	// Ji

#define ASKB_FN			0x4d	// red ALT.

class ASkeyboard {
public:
	static ASkeyboard* getInstance() {
		static ASkeyboard ask;
		return &ask;
	};
	bool begin(keyboard_notify* p = 0);
	void task(uint8_t hid_led_state = 0);

	void set_setup_fn(report_callback_t fn) { setup_callback = fn; }
	
	bool is_nicola_disabled() const;	// dip-sw pos == JIS.
	void setup_led(bool on);
	bool dedicated_oyayubi() const;

private:
	uint8_t kbd_led_state;	// hid keyboard led.
	uint8_t modifier_state;
	uint8_t last_modifier_state;
	bool fn_key_pressed;
	bool keyboard_locked;
	
	keyboard_notify* notify;
	
	ASkeyboard();
	void hid_convert(uint8_t k, bool pressed);
	void led_demo();
	report_callback_t setup_callback;
};
