/**
 * 
  BaseKeyboard.h Definitions of BaseKeyboard class for "Hobo-nicola keyboard adapter".
  Copyright (c) 2018 Takeshi Higasa
  
  This file is part of "Hobo-nicola keyboard adapter".

  "Hobo-nicola keyboard adapter" is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  "Hobo-nicola keyboard adapter" is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with "Hobo-nicola keyboard adapter".  If not, see <http://www.gnu.org/licenses/>.
  
*/

#include <Arduino.h>
#include "hbn_defines.h"

// following bytes have hid-id
#define HID_DIRECT_PREFIX	0x01

class BaseKeyboard {
public:
	BaseKeyboard();

	bool isScrLock() const;

	void sendReport(bool idle = false);
	void releaseAll();
	void report_press(uint8_t key, uint8_t mod);
	void report_release(uint8_t key, uint8_t mod, bool use_mod = false);
	bool isAllReleased() const;
	
	void stroke(uint8_t key, uint8_t mod);		// press and release. hid-code.
	void strokeChar(uint8_t c, uint8_t& mod);	// press and release. ascii-code.
	void send_PGM_string(const uint8_t* p);	// send pgm string.

	const bool isShiftPressed() const { return (modifier_keys & MODIFIER_HID_SHIFT_MASK) != 0; }

	void modifier_change(uint8_t before, uint8_t after);
	virtual void modifier_pressed(uint8_t mod) {}

	void idle();


	uint8_t get_hid_led_state() const;
	uint8_t get_hid_idle_value() const;

protected:
	unsigned long last_report_time;
	uint8_t modifier_keys;
	key_report_t report;

	uint16_t char_to_hid(uint8_t c);
	void append_hid_desc();
};
