#include "esp32_ble_hid_keyboard.h"

#if defined(__ESP32_BLE_HID_KEYBOARD__)

ESP32_BLE_HID_Keyaboard::ESP32_BLE_HID_Keyaboard() {
	connected = false;
	led_state = 0;  // set by hid output report.
	led_pin = 0xff;
    input = 0;
    output = 0;
	led_notify = 0;
}

void ESP32_BLE_HID_Keyaboard::set_kbd_led_state(uint8_t led) {
	led_state = led;
	if (led_notify)
		led_notify(led);

}

void ESP32_BLE_HID_Keyaboard::set_connected(bool f) { 
	connected = f; 
	if (led_pin != 0xff)
		digitalWrite(led_pin, f ? 1 : 0);
}

void ESP32_BLE_HID_Keyaboard::sendReport(key_report_t* p) {
	if (is_connected() && input) {
		input->setValue((uint8_t*)p, sizeof(key_report_t));
		input->notify();
	}
}

void ESP32_BLE_HID_Keyaboard::init_ble() {
	BLEDevice::init("ESP32_BLE_HOBO_NICOLA1");
//	BLEDevice::setPower(ESP_PWR_LVL_N12 );	// -12dbm minimum intensity.
	BLEDevice::setPower(ESP_PWR_LVL_P9  );	// +9dbm maximum intensity

	pServer = BLEDevice::createServer();
	BLEDevice::setEncryptionLevel(ESP_BLE_SEC_ENCRYPT);
	BLEDevice::setSecurityCallbacks(new SecurityCallback());
	pServer->setCallbacks(new ServerCallbacks());
	//	BLEDevice::setMTU(23);
	hid = new BLEHIDDevice(pServer);
	input = hid->inputReport(1);
	output = hid->outputReport(1);
	output->setCallbacks(new BleHidOutputReport());
	std::string name = "okiraku-camera";
	hid->manufacturer()->setValue(name);
	hid->pnp(0x02, 0xe502, 0xa111, 0x0210);
	hid->hidInfo(0x00,0x01);  // country == 0, flags == 1 ( providing wake-up signal to a HID host)
	hid->reportMap((uint8_t*)_hidReportDescriptor, sizeof(_hidReportDescriptor));
	hid->startServices();
	BLEAdvertising *pAdvertising = pServer->getAdvertising();
	pAdvertising->setAppearance(HID_KEYBOARD);
	pAdvertising->addServiceUUID(hid->hidService()->getUUID());
	pAdvertising->start();
	Serial.println("Start Advertising.");

	BLESecurity *pSecurity = new BLESecurity();
	pSecurity->setAuthenticationMode(ESP_LE_AUTH_REQ_SC_BOND);
//	pSecurity->setCapability(ESP_IO_CAP_IO);
	pSecurity->setCapability(ESP_IO_CAP_NONE);
	pSecurity->setInitEncryptionKey(ESP_BLE_ENC_KEY_MASK | ESP_BLE_ID_KEY_MASK);
}

void ServerCallbacks::onConnect(BLEServer* pServer) {
	ESP32_BLE_HID_Keyaboard::getInstance()->set_connected(true);
	Serial.printf("ServerCallbacks::onConnect()\n");
#if 1
//	BLEServer::getConnId(); public: ɂĂƁB
	esp_power_level_t tx_power = esp_ble_tx_power_get((esp_ble_power_type_t)pServer->getConnId());
	Serial.printf("ServerCallbacks::onConnect() connId = %d, tx_power = %d\n", pServer->getConnId(), tx_power);
#endif
};
  
void ServerCallbacks::onDisconnect(BLEServer* pServer) {
	ESP32_BLE_HID_Keyaboard::getInstance()->set_connected(false);
	Serial.println("ServerCallbacks::onDisconnect()");
}
  
void BleHidOutputReport::onRead(BLECharacteristic* me){
        Serial.printf("BleHidOutputReport::OnRead()\n");
}
void BleHidOutputReport::onWrite(BLECharacteristic* me){
	uint8_t* value = (uint8_t*)(me->getValue().c_str());
	if (value) {
		uint8_t led = *value;
		Serial.printf("BleHidOutputReport::onWrite() :  %02x\n", led);
		ESP32_BLE_HID_Keyaboard::getInstance()->set_kbd_led_state(led);
	}
}

uint8_t passKey = 0;
uint32_t SecurityCallback::onPassKeyRequest() {
	Serial.printf("onPassKeyRequest() return %d\n", passKey);
	return passKey;
}

void SecurityCallback::onPassKeyNotify(uint32_t pass_key){
	Serial.printf("onPassKeyNotify() notified = %d\n", pass_key);
	passKey = pass_key;
}

bool SecurityCallback::onSecurityRequest(){
	Serial.printf("onSecurityRequest() \n");
	return true;
}

// typedef uint8_t esp_link_key[ESP_BT_OCTET16_LEN];      /* Link Key */
void SecurityCallback::dump_auth_comp(esp_ble_auth_cmpl_t* p) {
	if (p) {
		BLEDevice::whiteListAdd(p->bd_addr);  // white list add.

		Serial.printf("remote BD_ADDR: %s\n", BLEAddress(p->bd_addr).toString().c_str());
		Serial.printf("address type = %d\n", p->addr_type);
		Serial.printf("key_type = %d\n", p->key_type);
		Serial.printf("dev_type = %d\n", p->dev_type);
		Serial.printf("esp_link_key : ");
		for(int i = 0; i < ESP_BT_OCTET16_LEN; i++)
			Serial.printf("%02X ", p->key[i]);
		Serial.println();
    }
}

void SecurityCallback::onAuthenticationComplete(esp_ble_auth_cmpl_t auth_cmpl){
	if (auth_cmpl.success) {
		dump_auth_comp(&auth_cmpl);
		BLECharacteristic* input = ESP32_BLE_HID_Keyaboard::getInstance()->getInputReport();
		if (input) {
			BLE2902 *p2902 = (BLE2902*)(input->getDescriptorByUUID((uint16_t)0x2902));
			if (p2902) {
				p2902->setNotifications(true);
			    Serial.printf("onAuthenticationComplete() enable input notification.\n");
			}
		}
	}
    Serial.printf("onAuthenticationComplete() pair status = %s\n", auth_cmpl.success ? "success" : "fail");
}

bool SecurityCallback::onConfirmPIN(uint32_t pin) {
	Serial.printf("onConfirmPIN() pin = %d\n", pin);
	return true;
}

#endif
