// esp32_ble_hid_keyboard.h

#if defined(ESP32)
#if !defined(__ESP32_BLE_HID_KEYBOARD__)
#define __ESP32_BLE_HID_KEYBOARD__

#include <Arduino.h>
#include "hbn_defines.h"

#include <BLEDevice.h>
#include <BLEServer.h>
#include <BLEUtils.h>
#include <BLE2902.h>
#include <BLEHIDDevice.h>



class ServerCallbacks: public BLEServerCallbacks {
	virtual void onConnect(BLEServer* pServer);
	virtual void onDisconnect(BLEServer* pServer);
};

class BleHidOutputReport : public BLECharacteristicCallbacks {
public:
	virtual void onRead(BLECharacteristic* me);
	virtual void onWrite(BLECharacteristic* me);
};

class SecurityCallback : public BLESecurityCallbacks {
public:
	virtual uint32_t onPassKeyRequest();
    virtual void onPassKeyNotify(uint32_t pass_key);
    virtual bool onSecurityRequest();
	virtual void onAuthenticationComplete(esp_ble_auth_cmpl_t auth_cmpl);
	virtual bool onConfirmPIN(uint32_t pin);
private:
  void dump_auth_comp(esp_ble_auth_cmpl_t* p);
};

class ESP32_BLE_HID_Keyaboard {
private:
	uint8_t led_state;
	uint8_t led_pin;
	bool connected;
	report_callback_t led_notify; 
	ESP32_BLE_HID_Keyaboard();
	

public:
	static ESP32_BLE_HID_Keyaboard* getInstance() {
		static ESP32_BLE_HID_Keyaboard instance;
		return &instance; 
	};

	void init_ble();
	void sendReport(key_report_t* p);
	void set_led_callback(report_callback_t fn) {
		led_notify= fn;
	};
	void set_kbd_led_state(uint8_t led);

	void set_connect_led_pin(uint8_t pin) { led_pin = pin; }
	void set_connected(bool f);
	bool is_connected() const { return connected; }

	BLECharacteristic* getInputReport() { return input; }


private:
	BLECharacteristic* input;
	BLECharacteristic* output;
	BLEHIDDevice* hid;
	BLEServer* pServer;

};
#endif
#endif		// ESP32


