/**
 * 
  usb_hobo_nicola.ino Main sketch of "Hobo-nicola keyboard adapter" (usb keyboard version).
  Copyright (c) 2018 Takeshi Higasa
  
  This file is part of "Hobo-nicola keyboard adapter".

  "Hobo-nicola keyboard adapter" is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  "Hobo-nicola keyboard adapter" is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with "Hobo-nicola keyboard adapter".  If not, see <http://www.gnu.org/licenses/>.

  version 1.2.1  Dec.10, 2018. support Compound USB device (PFU HHKB).
  
*/

#include "nicola_keyboard.h"
#include <hidboot.h>
#include <usbhub.h>
#include <SPI.h>

NicolaKeyboard hobo_nicola;

class KeyboardEvent : public KeyboardReportParser {
	USBHID *phid;

public:
	KeyboardEvent() { 
		phid = 0;
	};
	void Parse(USBHID *hid, bool is_rpt_id, uint8_t len, uint8_t *buf) {
		phid = hid;
		KeyboardReportParser::Parse(hid, is_rpt_id, len, buf);
	};
	
	void sync_led( uint8_t led ) {
		if (phid) phid->SetReport(0, 0, 2, 0, 1, &led);
	};
	
  protected:
	void OnControlKeysChanged(uint8_t before, uint8_t after) { hobo_nicola.modifier_change(before, after); };
	void OnKeyPressed(uint8_t key) {};
	void OnKeyDown  (uint8_t mod, uint8_t key) { hobo_nicola.key_press(key, mod);  };
	void OnKeyUp  (uint8_t mod, uint8_t key) { hobo_nicola.key_release(key, mod); };
};

USB Usb;
USBHub Hub1(&Usb);
USBHub Hub2(&Usb);
USBHub Hub3(&Usb);
USBHub Hub4(&Usb);
HIDBoot<USB_HID_PROTOCOL_KEYBOARD>    HidKeyboard(&Usb);
KeyboardEvent kbd;


void setup() {
	if (Usb.Init() == -1)
		hobo_nicola.error_blink();
	delay( 200 );
	HidKeyboard.SetReportParser(0, &kbd);
}

void loop() {
	Usb.Task();
	hobo_nicola.idle();
	kbd.sync_led(hobo_nicola.get_hid_led_state());
}
