// hoboNicola defines.
// hbn_defines.h

#include "Arduino.h"

#if !defined(__HOBO_NICOLA_DEFINES__)
#define __HOBO_NICOLA_DEFINES__


#define HBN_LIBRARY_VERSION	"1_2_1"


#define REPORT_KEYS 6
typedef struct {
	uint8_t modifiers;
	uint8_t reserved;
	uint8_t keys[REPORT_KEYS];
} key_report_t;

typedef void (*report_callback_t)(uint8_t);

class keyboard_notify {
public:
	virtual void key_pressed(uint8_t k, uint8_t mod) {}; // notify key pressed with hid usage id.
	virtual void key_released(uint8_t k, uint8_t mod) {}; // notify key released with hid usage id.
	virtual void modifier_changed(uint8_t before, uint8_t after) {}; // notify any modifier-key is changed.
	virtual void raw_key_event(uint8_t scan, bool released = false) {};  // notify all key event with raw scancode.
	virtual void key_released_all() {}; // all keys are released.
};

#if 1
// ASkeyboard OYAYUBI-keys
#define ASKB_LEFT_OYAYUBI	0x72		// HID Keyboard F23
#define ASKB_RIGHT_OYAYUBI	0x73		// HID Keyboard F24
#endif

#define MODIFIRE_HID_CTRL_L		0x01
#define MODIFIRE_HID_CTRL_R		0x10
#define MODIFIRE_HID_CTRL_MASK		0x11
#define MODIFIRE_HID_SHIFT_L	0x02
#define MODIFIRE_HID_SHIFT_R	0x20
#define MODIFIER_HID_SHIFT_MASK	0x22
#define MODIFIRE_HID_ALT_L		0x04
#define MODIFIRE_HID_ALT_R		0x40
#define MODIFIRE_HID_ALT_MASK	0x44
#define MODIFIRE_HID_GUI_L		0x08
#define MODIFIRE_HID_GUI_R		0x80
#define MODIFIRE_HID_GUI_MASK	0x88


// Keyboard LEDs
#define KEY_LED_NUMLOCK   1
#define KEY_LED_CAPSLOCK  2
#define KEY_LED_SCRLOCK   4
#define KEY_LED_COMPOSE   8
#define KEY_LED_KANA    0x10

#if !defined(ESP32)	// Pro Micro only.
#define _HBN_NICOLA_LED	LED_BUILTIN_RX
#define _HBN_ERROR_LED	LED_BUILTIN_TX
#define HBN_NICOLA_LED(a) digitalWrite(_HBN_NICOLA_LED, a ? 0 : 1)
#define HBN_ERROR_LED(a) digitalWrite(_HBN_ERROR_LED, a ? 0 : 1)
#define TOGGLE_HBN_NICOLA_LED() digitalWrite(_HBN_NICOLA_LED, digitalRead(_HBN_NICOLA_LED) ^ 1)
#define TOGGLE_HBN_ERROR_LED() digitalWrite(_HBN_ERROR_LED, digitalRead(_HBN_ERROR_LED) ^ 1)

#else
// LEDの接続が必要。
#define _HBN_NICOLA_LED	16
#define _HBN_ERROR_LED	0
#define HBN_NICOLA_LED(a)
#define HBN_ERROR_LED(a) 
#define TOGGLE_HBN_NICOLA_LED()
#endif

// PROGMEM related.
#if defined(ESP32)
#define PROGMEM
#endif

#if defined(ESP32)
#define PGM_READ_ADDR(a)	(uint8_t*)a
#define PGM_READ_BYTE(a)	*a
#else
#define PGM_READ_ADDR(a)	(uint8_t*)pgm_read_word(&a)
#define PGM_READ_BYTE(a)	(uint8_t)pgm_read_byte(a)
#endif


// HID Report Descriptor.
static const uint8_t _hidReportDescriptor[] PROGMEM = {
	0x05, 0x01, // USAGE_PAGE (Generic Desktop)
	0x09, 0x06, // USAGE (Keyboard)
	0xa1, 0x01, // COLLECTION (Application)
#if !defined(ESP32)	
	0x85, 0x02, //   REPORT_ID (2) --- Mouse.cppがID==1。
#else
	0x85, 0x01, //   REPORT_ID (1)
#endif
	0x05, 0x07, //   USAGE_PAGE (usage = keyboard page)
	// モデファイヤキー(修飾キー)
	0x19, 0xe0, //     USAGE_MINIMUM (左CTRLが0xe0)
	0x29, 0xe7, //     USAGE_MAXIMUM (右GUIが0xe7)
	0x15, 0x00, //   LOGICAL_MINIMUM (0)
	0x25, 0x01, //   LOGICAL_MAXIMUM (1)
	0x95, 0x08, //   REPORT_COUNT (8) 全部で8つ(左右4つずつ)。
	0x75, 0x01, //   REPORT_SIZE (1) 各修飾キーにつき1ビット
	0x81, 0x02, //   INPUT (Data,Var,Abs) 8ビット長のInputフィールド(変数)が1つ。
	// 予約フィールド
	0x95, 0x01, //   REPORT_COUNT (1)
	0x75, 0x08, //   REPORT_SIZE (8) 1ビットが8つ。
	0x81, 0x01, //   INPUT (Cnst,Var,Abs)
	// LED状態のアウトプット
	0x95, 0x05, //   REPORT_COUNT (5) 全部で5つ。
	0x75, 0x01, //   REPORT_SIZE (1)  各LEDにつき1ビット
	0x05, 0x08, //   USAGE_PAGE (LEDs)
	0x19, 0x01, //     USAGE_MINIMUM (1) (NumLock LEDが1)
	0x29, 0x05, //     USAGE_MAXIMUM (5) (KANA LEDが5)
	0x91, 0x02, //   OUTPUT (Data,Var,Abs) // LED report
	// LEDレポートのパディング
	0x95, 0x01, //   REPORT_COUNT (1)
	0x75, 0x03, //   REPORT_SIZE (3)　　残りの3ビットを埋める。
	0x91, 0x01, //   OUTPUT (Cnst,Var,Abs) // padding 
	// 押下情報のインプット
	0x95, 0x06, //   REPORT_COUNT (6) 全部で6つ。
	0x75, 0x08, //   REPORT_SIZE (8) おのおの8ビットで表現
	0x15, 0x00, //   LOGICAL_MINIMUM (0) キーコードの範囲は、
	0x25, 0xdd, //   LOGICAL_MAXIMUM (221) 0～221(0xdd)まで

	0x05, 0x07, //   USAGE_PAGE (Keyboard)
	0x19, 0x00, //     USAGE_MINIMUM (0はキーコードではない)
	0x29, 0xdd, //     USAGE_MAXIMUM (Keypad Hexadecimalまで)
	0x81, 0x00, //   INPUT (Data,Ary,Abs)
	0xc0    // END_COLLECTION
};

#endif
