/**
 * 
  nicola_keyboard.h Definitions of NicolaKeyboard class for "Hobo-nicola keyboard adapter".
  Copyright (c) 2018 Takeshi Higasa
  
  This file is part of "Hobo-nicola keyboard adapter".

  "Hobo-nicola keyboard adapter" is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  "Hobo-nicola keyboard adapter" is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with "Hobo-nicola keyboard adapter".  If not, see <http://www.gnu.org/licenses/>.
  
*/

#include "BaseKeyboard.h"
#include "nicola_table.h"

#define isNicola() isScrLock()

typedef enum {  // ステートマシンに与えるイベント
	Init = 0,
	Moji_pressed,
	Oyayubi_pressed,
	Key_released,
	Time_out,
	All_off,
	Key_repeat
} nicola_event_t;

// settings in EEPROM
#define SPC_TO_LEFT		1
#define SPC_TO_RIGHT	2
#define SPC_TO_OYAYUBI	3
#define OUTPUT_IMMEDIATE_ON		4
#define EISU_TO_SCR_OFF	8
#define KANA_TO_SCR_ON	0x10
#define KANJI_TOGGLE_SCRLOCK	0x20
#define SELF_REPEAT		0x80

class NicolaKeyboard : public BaseKeyboard {
	uint32_t settings;	// current copy of eeprom.
	bool disable_nicola;
	bool decicated_oyakeys;
public:
	NicolaKeyboard();

	void key_press(uint8_t k, uint8_t mod);
	void key_release(uint8_t k,uint8_t mod);
	virtual void modifier_pressed(uint8_t mod);
	void output() ;
	void idle();
	void error_blink(int period = 500);
	void set_disable_nicola(bool f = true) { disable_nicola = f; }

	enum {
		Initial_State = 0,		// 初期状態
		Oyayubi_State,			// 親指キー押下状態
		Character_State,		// 文字キー押下状態
		Char_Oya_State,			// 文字キー押下中の親指キー押下状態
		Repeat_State			// リピート中状態
	} state;

// settings.
	bool is_immediate_output() const { return (bool)(settings & OUTPUT_IMMEDIATE_ON); }
	bool is_spc_to_left() const { return (bool)(settings & SPC_TO_LEFT); }
	bool is_spc_to_right() const { return (bool)(settings & SPC_TO_RIGHT); }
	bool is_spc_to_none() const { return (bool) (settings & SPC_TO_OYAYUBI) == 0; }
	bool is_self_repeat() const { return (bool)(settings & SELF_REPEAT); }
	
	bool is_eisu_to_scr_off() const { return (bool)(settings & EISU_TO_SCR_OFF); }
	bool is_kana_to_scr_on() const { return (bool)(settings & KANA_TO_SCR_ON); }
	bool is_kanji_toggle_scrlock() const { return (bool)(settings & KANJI_TOGGLE_SCRLOCK); }

	void enter_setup_mode(bool f = true) { setup_mode = f; }
	bool is_setup_mode() const { return setup_mode; }
	void has_dedicated_oyakeys(bool f = true) { decicated_oyakeys = f; }

private:
	void nicola_state(nicola_event_t e, uint16_t param = 0);
	void moji_set(uint16_t param);
	void oyayubi_set(uint16_t param);

	uint16_t convert_keycode(uint8_t k);
	uint8_t pre_convert(uint8_t k);
	const uint8_t* get_output_data(uint16_t moji, uint16_t oyayubi);  // return pgm address.

	unsigned long event_time;
	unsigned long moji_time;
	unsigned long oyayubi_time;
	
	uint16_t oyayubi;
	uint16_t moji;
	uint16_t repeat_moji;
	uint16_t repeat_oyayubi;

	uint16_t repeat_key;
	unsigned long repeat_time;

	const unsigned long e_charTime = 200;
	const unsigned long e_oyaTime = 200;
	const unsigned long e_nicolaTime = 80;

	const unsigned long repeat_delay = 250;
	const unsigned long repeat_interval = 80;

	bool immediate;

	void immediate_output(uint16_t moji);
	void sendBS() ;

	bool setup_mode;
	void setup_options(uint8_t nid);
	
	const uint32_t load_settings();
	void save_settings(uint32_t new_set);
	void show_setting();
	
};

