/**
 * 
  nicola_table.cpp  Code tables and related functions of "Hobo-nicola keyboard adapter".
  Copyright (c) 2018 Takeshi Higasa
  
  This file is part of "Hobo-nicola keyboard adapter".

  "Hobo-nicola keyboard adapter" is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  "Hobo-nicola keyboard adapter" is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with "Hobo-nicola keyboard adapter".  If not, see <http://www.gnu.org/licenses/>.
  
*/

#include "hbn_defines.h"
#include "nicola_keyboard.h"

// hid usage id(hid) --> internal code(nid)
//	1 2 3 4 5 6 7 8 9 0 - ^ |
//  q w e r t y u i o p @ [
//  a s d f g h j k l ; : ]
//  z x c v b n m , . / _

//static static const uint8_t PROGMEM hid_to_nid[] = {
static const uint8_t hid_to_nid[] = {
  0x1e, 0x1f, 0x20, 0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27, 0x2d, 0x2e, 0x89,	// E段 1～0x0d
  0x14, 0x1a, 0x08, 0x15, 0x17, 0x1c, 0x18, 0x0c, 0x12, 0x13, 0x2f, 0x30,		// D段 0x0e～0x19
  0x04, 0x16, 0x07, 0x09, 0x0a, 0x0b, 0x0d, 0x0e, 0x0f, 0x33, 0x34, 0x32,		// C段 0x1a～0x25
  0x1d, 0x1b, 0x06, 0x19, 0x05, 0x11, 0x10, 0x36, 0x37, 0x38, 0x87				// B段 0x26～0x30
};

#define NUM_NICOLA_KEYS	sizeof(hid_to_nid)

// hid -> nid. otherwise return 0;
uint16_t NicolaKeyboard::convert_keycode(uint8_t k) {
	switch(k) {
	case HID_SPACE:
		if (is_spc_to_left())
			return k << 8 | NID_LEFT_OYAYUBI;
		else if (is_spc_to_right())
			return k << 8 | NID_RIGHT_OYAYUBI;
		break;
	case HID_MUHENKAN:
		if (!decicated_oyakeys)
			return k << 8 | NID_LEFT_OYAYUBI;
		break;
	case HID_HENKAN:
		if (!decicated_oyakeys)
			return k << 8 | NID_RIGHT_OYAYUBI;
		break;
	case HID_PAUSE:
		return k << 8 | NID_SETUP_KEY;
#if 1
	case ASKB_LEFT_OYAYUBI:
		return NID_LEFT_OYAYUBI;
	case ASKB_RIGHT_OYAYUBI:
		return NID_RIGHT_OYAYUBI;
#endif
	default:		
		for(uint8_t i = 0; i < NUM_NICOLA_KEYS; i++) {
			// if (pgm_read_byte(&hid_to_nid[i] == k)
			if (hid_to_nid[i] == k)
				return k << 8 | (i + 1);
		}
	}
	return 0;  
}

// isNicola()のときに、
// 後退キーなど状態遷移に関与しないがコード変換が必要なキー。
// k : hid usage id.
uint8_t NicolaKeyboard::pre_convert(uint8_t k) {
	if (k == HID_COLON)            // C11
		return HID_BACKSPACE;
	else if (k == HID_RIGHT_SQUARE_BRACKET)       // C12
		return HID_ESCAPE;
	return 0;
}
//
// 出力する文字列（ローマ字シーケンス）
// 先頭バイトが0x01ならば引き続くバイトはHIDコード。
// 
static const uint8_t empty_str[] PROGMEM = "";

// output chars (roma-ji)
// single stroke.
static const uint8_t e_1[] PROGMEM = "1";
static const uint8_t e_2[] PROGMEM = "2";
static const uint8_t e_3[] PROGMEM = "3";
static const uint8_t e_4[] PROGMEM = "4";
static const uint8_t e_5[] PROGMEM = "5";
static const uint8_t e_6[] PROGMEM = "6";
static const uint8_t e_7[] PROGMEM = "7";
static const uint8_t e_8[] PROGMEM = "8";
static const uint8_t e_9[] PROGMEM = "9";
static const uint8_t e_10[] PROGMEM = "0";
static const uint8_t e_11[] PROGMEM = "-";
static const uint8_t e_12[] PROGMEM = "|";
static const uint8_t e_13[] PROGMEM = "\\";
static const uint8_t d_1[] PROGMEM = ".";	//句点
static const uint8_t d_2[] PROGMEM = "ka";
static const uint8_t d_3[] PROGMEM = "ta";
static const uint8_t d_4[] PROGMEM = "ko";
static const uint8_t d_5[] PROGMEM = "sa";
static const uint8_t d_6[] PROGMEM = "ra";
static const uint8_t d_7[] PROGMEM = "ti";
static const uint8_t d_8[] PROGMEM = "ku";
static const uint8_t d_9[] PROGMEM = "tu";
static const uint8_t d_10[] PROGMEM = ",";	// カンマ
static const uint8_t d_11[] PROGMEM = ",";	// 読点
static const uint8_t d_12[] PROGMEM = ":";
static const uint8_t c_1[] PROGMEM = "u";
static const uint8_t c_2[] PROGMEM = "si";
static const uint8_t c_3[] PROGMEM = "te";
static const uint8_t c_4[] PROGMEM = "ke";
static const uint8_t c_5[] PROGMEM = "se";
static const uint8_t c_6[] PROGMEM = "ha";
static const uint8_t c_7[] PROGMEM = "to";
static const uint8_t c_8[] PROGMEM = "ki";
static const uint8_t c_9[] PROGMEM = "i";
static const uint8_t c_10[] PROGMEM = "nn";
// static const uint8_t c_11[] PROGMEM = "";	// 後退
// static const uint8_t c_12[] PROGMEM = "";  // 取消
#define c_11 empty_str
#define c_12 empty_str
static const uint8_t b_1[] PROGMEM = ".";	// ピリオド
static const uint8_t b_2[] PROGMEM = "hi";
static const uint8_t b_3[] PROGMEM = "su";
static const uint8_t b_4[] PROGMEM = "fu";
static const uint8_t b_5[] PROGMEM = "he";
static const uint8_t b_6[] PROGMEM = "me";
static const uint8_t b_7[] PROGMEM = "so";
static const uint8_t b_8[] PROGMEM = "ne";
static const uint8_t b_9[] PROGMEM = "ho";
static const uint8_t b_10[] PROGMEM = "/";	// 中黒
static const uint8_t b_11[] PROGMEM = "\\";

// single stroke with SHIFT
static const uint8_t e_1S[] PROGMEM = "!";
static const uint8_t e_2S[] PROGMEM = "\"";
static const uint8_t e_3S[] PROGMEM = "#";
static const uint8_t e_4S[] PROGMEM = "$";	// 「
static const uint8_t e_5S[] PROGMEM = "%";	// 」
static const uint8_t e_6S[] PROGMEM = "&";	// [
static const uint8_t e_7S[] PROGMEM = "'";	// ]
static const uint8_t e_8S[] PROGMEM = "<";
static const uint8_t e_9S[] PROGMEM = ">";
#define e_10S empty_str
static const uint8_t e_11S[] PROGMEM = "=";
static const uint8_t e_12S[] PROGMEM = "~";
static const uint8_t e_13S[] PROGMEM = "|";
#define d_1S  empty_str
#define d_2S  empty_str
#define d_3S  empty_str
#define d_4S  empty_str
#define d_5S  empty_str
#define d_6S  empty_str
#define d_7S  empty_str
#define d_8S  empty_str
#define d_9S  empty_str
#define d_10S  empty_str
static const uint8_t d_11S[] PROGMEM = "`";	// back-q
static const uint8_t d_12S[] PROGMEM = "*";
#define c_1S  empty_str
#define c_2S  empty_str
static const uint8_t c_3S[] PROGMEM = {HID_DIRECT_PREFIX, HID_LEFT_ARROW, 0};		// D  ←
static const uint8_t c_4S[] PROGMEM = {HID_DIRECT_PREFIX, HID_DOWN_ARROW, 0};		// F  ↓
#define c_5S  empty_str
static const uint8_t c_6S[] PROGMEM = "pa";
static const uint8_t c_7S[] PROGMEM = {HID_DIRECT_PREFIX, HID_UP_ARROW, 0};			// J  ↑
static const uint8_t c_8S[] PROGMEM = {HID_DIRECT_PREFIX, HID_RIGHT_ARROW, 0};			// K  →
#define c_9S  empty_str
static const uint8_t c_10S[] PROGMEM = "+";
#define c_11S  empty_str
#define c_12S  empty_str
#define b_1S  empty_str
static const uint8_t b_2S[] PROGMEM = "pi";
#define b_3S  empty_str
static const uint8_t b_4S[] PROGMEM = "pu";
static const uint8_t b_5S[] PROGMEM = "pe";		// B
static const uint8_t b_6S[] PROGMEM = {HID_DIRECT_PREFIX, HID_DELETE, 0};			// N Delete
static const uint8_t b_7S[] PROGMEM = {HID_DIRECT_PREFIX, HID_ENTER, 0};			// M Enter
#define b_8S  empty_str
static const uint8_t b_9S[] PROGMEM = "po";
static const uint8_t b_10S[] PROGMEM = "?";
static const uint8_t b_11S[] PROGMEM = "_";

// simultaneous stroke with Left Thumb-key
static const uint8_t e_1L[] PROGMEM = "?";
static const uint8_t e_2L[] PROGMEM = "/";
static const uint8_t e_3L[] PROGMEM = "~";
static const uint8_t e_4L[] PROGMEM = "[";
static const uint8_t e_5L[] PROGMEM = "]";
static const uint8_t e_6L[] PROGMEM = "[";
static const uint8_t e_7L[] PROGMEM = "]";
static const uint8_t e_8L[] PROGMEM = "(";
static const uint8_t e_9L[] PROGMEM = ")";
static const uint8_t e_10L[] PROGMEM = "{";
static const uint8_t e_11L[] PROGMEM = "}";
static const uint8_t e_12L[] PROGMEM = "\\";
#define e_13L  empty_str
static const uint8_t d_1L[] PROGMEM = "xa";	// ぁ
static const uint8_t d_2L[] PROGMEM = "e";	// え
static const uint8_t d_3L[] PROGMEM = "ri";	// り
static const uint8_t d_4L[] PROGMEM = "xya";	// ゃ	
static const uint8_t d_5L[] PROGMEM = "re";	// れ
static const uint8_t d_6L[] PROGMEM = "pa";	// ぱ
static const uint8_t d_7L[] PROGMEM = "di";	// ぢ
static const uint8_t d_8L[] PROGMEM = "gu";	// ぐ
static const uint8_t d_9L[] PROGMEM = "du";	// づ
static const uint8_t d_10L[] PROGMEM = "pi";	// ぴ
static const uint8_t d_11L[] PROGMEM = "@";	// 
static const uint8_t d_12L[] PROGMEM = "*";	// *
static const uint8_t c_1L[] PROGMEM = "wo";	// を
static const uint8_t c_2L[] PROGMEM = "a";	// あ
static const uint8_t c_3L[] PROGMEM = "na";	// な
static const uint8_t c_4L[] PROGMEM = "xyu";	// ゅ
static const uint8_t c_5L[] PROGMEM = "mo";	// も
static const uint8_t c_6L[] PROGMEM = "ba";	// ば
static const uint8_t c_7L[] PROGMEM = "do";	// ど
static const uint8_t c_8L[] PROGMEM = "gi";	// ぎ
static const uint8_t c_9L[] PROGMEM = "po";	// ぽ
#define c_10L  empty_str
#define c_11L  empty_str
#define c_12L  empty_str
static const uint8_t b_1L[] PROGMEM = "xu";	// ぅ
static const uint8_t b_2L[] PROGMEM = "-";	// ー 音引き
static const uint8_t b_3L[] PROGMEM = "ro";	// ろ
static const uint8_t b_4L[] PROGMEM = "ya";	// や
static const uint8_t b_5L[] PROGMEM = "xi";	// ぃ
static const uint8_t b_6L[] PROGMEM = "pu";	// ぷ
static const uint8_t b_7L[] PROGMEM = "zo";	// ぞ
static const uint8_t b_8L[] PROGMEM = "pe";	// ぺ
static const uint8_t b_9L[] PROGMEM = "bo";	// ぼ
#define b_10L  empty_str
#define b_11L  empty_str

// simultaneous stroke with Right Thumb-key
static const uint8_t e_1R[] PROGMEM = "?";
static const uint8_t e_2R[] PROGMEM = "/";
static const uint8_t e_3R[] PROGMEM = "~";
static const uint8_t e_4R[] PROGMEM = "[";
static const uint8_t e_5R[] PROGMEM = "]";
static const uint8_t e_6R[] PROGMEM = "[";
static const uint8_t e_7R[] PROGMEM = "]";
static const uint8_t e_8R[] PROGMEM = "(";
static const uint8_t e_9R[] PROGMEM = ")";
static const uint8_t e_10R[] PROGMEM = "{";
static const uint8_t e_11R[] PROGMEM = "}";
static const uint8_t e_12R[] PROGMEM = "\\";
#define e_13R  empty_str
#define d_1R  empty_str
static const uint8_t d_2R[] PROGMEM = "ga";	// が
static const uint8_t d_3R[] PROGMEM = "da";	// だ
static const uint8_t d_4R[] PROGMEM = "go";	// ご
static const uint8_t d_5R[] PROGMEM = "za";	// ざ
static const uint8_t d_6R[] PROGMEM = "yo";	// よ
static const uint8_t d_7R[] PROGMEM = "ni";	// に
static const uint8_t d_8R[] PROGMEM = "ru";	// る
static const uint8_t d_9R[] PROGMEM = "ma";	// ま
static const uint8_t d_10R[] PROGMEM = "xe";	// ぇ
static const uint8_t d_11R[] PROGMEM = "@";	// 
static const uint8_t d_12R[] PROGMEM = "*";	// 
static const uint8_t c_1R[] PROGMEM = "vu";	// ヴ
static const uint8_t c_2R[] PROGMEM = "ji";	// じ
static const uint8_t c_3R[] PROGMEM = "de";	// で
static const uint8_t c_4R[] PROGMEM = "ge";	// げ
static const uint8_t c_5R[] PROGMEM = "ze";	// ぜ
static const uint8_t c_6R[] PROGMEM = "mi";	// み
static const uint8_t c_7R[] PROGMEM = "o";	// お
static const uint8_t c_8R[] PROGMEM = "no";	// の
static const uint8_t c_9R[] PROGMEM = "xyo";	// ょ
static const uint8_t c_10R[] PROGMEM = "xtu";	// っ
#define c_11R  empty_str
#define c_12R  empty_str
#define b_1R  empty_str
static const uint8_t b_2R[] PROGMEM = "bi";	// び
static const uint8_t b_3R[] PROGMEM = "zu";	// ず
static const uint8_t b_4R[] PROGMEM = "bu";	// ぶ
static const uint8_t b_5R[] PROGMEM = "be";	// べ
static const uint8_t b_6R[] PROGMEM = "nu";	// ぬ
static const uint8_t b_7R[] PROGMEM = "yu";	// ゆ
static const uint8_t b_8R[] PROGMEM = "mu";	// む
static const uint8_t b_9R[] PROGMEM = "wa";	// わ
static const uint8_t b_10R[] PROGMEM = "xo";	// ぉ
#define b_11R  empty_str

const uint8_t* const output_single[] PROGMEM = {
	e_1, e_2, e_3, e_4, e_5, e_6, e_7, e_8, e_9, e_10, e_11, e_12, e_13,
	d_1, d_2, d_3, d_4, d_5, d_6, d_7, d_8, d_9, d_10, d_11, d_12,
	c_1, c_2, c_3, c_4, c_5, c_6, c_7, c_8, c_9, c_10, c_11, c_12,
	b_1, b_2, b_3, b_4, b_5, b_6, b_7, b_8, b_9, b_10, b_11
};

const uint8_t* const output_left[] PROGMEM = {
	e_1L, e_2L, e_3L, e_4L, e_5L, e_6L, e_7L, e_8L, e_9L, e_10L, e_11L, e_12L, e_13L,
	d_1L, d_2L, d_3L, d_4L, d_5L, d_6L, d_7L, d_8L, d_9L, d_10L, d_11L, d_12L,
	c_1L, c_2L, c_3L, c_4L, c_5L, c_6L, c_7L, c_8L, c_9L, c_10L, c_11L, c_12L,
	b_1L, b_2L, b_3L, b_4L, b_5L, b_6L, b_7L, b_8L, b_9L, b_10L, b_11L
};

const uint8_t* const output_right[] PROGMEM = {
	e_1R, e_2R, e_3R, e_4R, e_5R, e_6R, e_7R, e_8R, e_9R, e_10R, e_11R, e_12R, e_13R,
	d_1R, d_2R, d_3R, d_4R, d_5R, d_6R, d_7R, d_8R, d_9R, d_10R, d_11R, d_12R,
	c_1R, c_2R, c_3R, c_4R, c_5R, c_6R, c_7R, c_8R, c_9R, c_10R, c_11R, c_12R,
	b_1R, b_2R, b_3R, b_4R, b_5R, b_6R, b_7R, b_8R, b_9R, b_10R, b_11R
};

const uint8_t* const output_shift[] PROGMEM = {
	e_1S, e_2S, e_3S, e_4S, e_5S, e_6S, e_7S, e_8S, e_9S, e_10S, e_11S, e_12S, e_13S,
	d_1S, d_2S, d_3S, d_4S, d_5S, d_6S, d_7S, d_8S, d_9S, d_10S, d_11S, d_12S,
	c_1S, c_2S, c_3S, c_4S, c_5S, c_6S, c_7S, c_8S, c_9S, c_10S, c_11S, c_12S,
	b_1S, b_2S, b_3S, b_4S, b_5S, b_6S, b_7S, b_8S, b_9S, b_10S, b_11S
};

// output support
// 出力すべき文字列のアドレスを返す。ATmegaならば文字列はPGM内にある。
// 
const uint8_t*  NicolaKeyboard::get_output_data(uint16_t moji, uint16_t oyayubi) {
	uint8_t n = LOWBYTE(moji) ;
	if (n < 1 || n > NUM_NICOLA_KEYS)
		return 0;
	n--;
	uint8_t* address = 0;
	if (oyayubi == 0)
		address = (uint8_t*)PGM_READ_ADDR(output_single[n]);
	else if (LOWBYTE(oyayubi) == NID_LEFT_OYAYUBI)
		address = (uint8_t*)PGM_READ_ADDR(output_left[n]);
	else if (LOWBYTE(oyayubi) == NID_RIGHT_OYAYUBI)
		address = (uint8_t*)PGM_READ_ADDR(output_right[n]);
	else if (LOWBYTE(oyayubi) == NID_SHIFT_KEY)
		address = (uint8_t*)PGM_READ_ADDR(output_shift[n]);
	if (address && PGM_READ_BYTE(address) == 0)
		return 0;
	return (const uint8_t*)address;    
}

// ------------------- setting related.
#include <EEPROM.h>
#define EEPROM_ADDR	0
const uint32_t NicolaKeyboard::load_settings() {
	uint32_t a = 0;
	uint32_t b = 0;
	
	EEPROM.get(EEPROM_ADDR, a);
	EEPROM.get(EEPROM_ADDR + 4, b);
	if (a != ~b) {
		a = 0;
		EEPROM.put(EEPROM_ADDR, a);
		EEPROM.put(EEPROM_ADDR + 4, ~a);
	}
	return a;
}

void NicolaKeyboard::save_settings(uint32_t new_set) {
	if (new_set == settings)
		return ;
	EEPROM.put(EEPROM_ADDR, new_set);
	EEPROM.put(EEPROM_ADDR + 4, ~new_set);
	settings = new_set;
#if defined(ESP32)
	EEPROM.commit();
#endif	
}

//
//	output settings as keyboard inputs.
//
static const uint8_t set_title[] PROGMEM =	"CURRENT SETTINGS OF HOBO_NICOLA";

static const uint8_t set_1[] PROGMEM =		"1 : SPC -> LEFT OYAYUBI";
static const uint8_t set_2[] PROGMEM =		"2 : SPC -> RIGHT OYAYUBI";
static const uint8_t set_3[] PROGMEM =		"3 : SPC -> NONE";
#define set_4 empty_str
static const uint8_t set_5[] PROGMEM =		"5 : REPEAT ON";
#define set_6 empty_str
static const uint8_t set_7[] PROGMEM =		"7 : EISU -> SCRLOCK OFF";
static const uint8_t set_8[] PROGMEM =		"8 : HIRAGANA -> SCRLOCK ON";
static const uint8_t set_9[] PROGMEM =		"9 : KANJI -> TOGGLE SCRLOCK";
static const uint8_t set_10[] PROGMEM = 	"0 : IMMEDIATE OUTPUT";
#define set_11	empty_str
#define set_12	empty_str
static const uint8_t set_done[] PROGMEM =  "VERSION  : " HBN_LIBRARY_VERSION "...";


const uint8_t* const output_settings[] PROGMEM = {
	set_title,
	set_1, set_2, set_3,
	set_4, set_5, set_6, 
	set_7, set_8, 
	set_9,
	set_10, 
	set_11,
	set_12, 
	set_done
};

void NicolaKeyboard::show_setting() {
	stroke(HID_ENTER, 0);
	for(uint8_t i = 0; i < sizeof(output_settings) / sizeof(uint8_t*); i++) {
		uint8_t* address = PGM_READ_ADDR(output_settings[i]);
		if (address) {
			if (pgm_read_byte(address)) {
				uint8_t mod = modifier_keys;
				bool f = false;
				switch(i) {
				case 1:
					f = is_spc_to_left();
					break;
				case 2:
					f = is_spc_to_right();
					break;
				case 3:
					f = is_spc_to_none();
					break;
				case 5:
					f = is_self_repeat();
					break;
				case 7:
					f = is_eisu_to_scr_off();
					break;
				case 8:
					f = is_kana_to_scr_on();
					break;
				case 9:
					f = is_kanji_toggle_scrlock();
					break;
					
				case 10:
					f = is_immediate_output();
					break;
				default:
					break;
				}
				if (i > 0 && i < (sizeof(output_settings) / sizeof(uint8_t*)) - 1)
					strokeChar(f ? '*' : ' ', mod);
				send_PGM_string((const uint8_t*)address);
				stroke(HID_ENTER, 0);
				stroke(HID_ENTER, 0);
			}
		}
	}
	stroke(HID_ENTER, 0);
	stroke(HID_ENTER, 0);
}

void  NicolaKeyboard::setup_options(uint8_t nid) {
	uint8_t new_settings = settings;
	switch(nid) {
	case NID_YEN:
		show_setting();
		break;
	case NID_0:
		if (is_immediate_output())
			new_settings &= ~OUTPUT_IMMEDIATE_ON;
		else
			new_settings |= OUTPUT_IMMEDIATE_ON;
		break;
	case NID_1:
		new_settings |= SPC_TO_LEFT;
		new_settings &= ~SPC_TO_RIGHT;
		break;
	case NID_2:
		new_settings |= SPC_TO_RIGHT;
		new_settings &= ~SPC_TO_LEFT;
		break;
	case NID_3:
		new_settings &= ~SPC_TO_OYAYUBI;
		break;
	case NID_5:
		if (is_self_repeat())
			new_settings &= ~SELF_REPEAT;
		else
			new_settings |= SELF_REPEAT;
		break;
	case NID_7:
		if (is_eisu_to_scr_off())
			new_settings &= ~EISU_TO_SCR_OFF;
		else
			new_settings |= EISU_TO_SCR_OFF;
		break;
	case NID_8:
		if (is_kana_to_scr_on())
			new_settings &= ~KANA_TO_SCR_ON;
		else
			new_settings |= KANA_TO_SCR_ON;
		break;
	case NID_9:
		if (is_kanji_toggle_scrlock())
			new_settings &= ~KANJI_TOGGLE_SCRLOCK;
		else
			new_settings |= KANJI_TOGGLE_SCRLOCK;
		break;
	}
	save_settings(new_settings);
	setup_mode = false;
}

